import time

import boto3
import numpy as np
import pandas as pd
import requests
from logbook import Logger, DEBUG

from catalyst import run_algorithm
from catalyst.api import symbol, record
from catalyst.exchange.exchange_errors import SymbolNotFoundOnExchange
from catalyst.exchange.utils.stats_utils import get_pretty_stats

algo_namespace = 'arbitrage_poloniex_bitfinex_bittrex'

log = Logger(algo_namespace)
log.level = DEBUG

s3 = boto3.resource('s3')


def initialize(context):
    context.signals = []

    context.hourly_orderbook_data = []
    context.current_hour = None

    context.num_opportunities = 0
    context.max_percent = 0

    context.targets = dict()

    context.targets['eth_btc'] = \
        dict([(1, 1), (1.5, 1.1), (1.75, 1.2), (2, 1.5), (3, 2)])

    context.targets['ltc_btc'] = \
        dict([(2, 1.0), (5, 1.5), (10, 2.0), (40, 2.5), (60, 3.0)])

    context.targets['xmr_btc'] = \
        dict([(1, -1), (3, 1.5), (8, -1), (30, 2.5), (40, 3.0)])

    context.targets['omg_btc'] = \
        dict([(50, 1.0), (100, 1.5), (200, 2.0), (400, 2.5), (600, 3.0)])

    context.targets['neo_btc'] = \
        dict([(2, 1.0), (5, 1.5), (10, 2.0), (40, 2.5), (60, 3.0)])

    # context.targets['dash_btc'] = \
    #     dict([(1, 1.0), (2, 1.5), (4, 2.0), (8, 2.5), (12, 3.0)])

    context.targets['etc_btc'] = \
        dict([(1, 1.0), (3, 1.5), (8, 2.0), (30, 2.5), (40, 3.0)])
    #
    # context.targets['zec_btc'] = \
    #     dict([(1, 1.0), (2, 1.5), (4, 2.0), (8, 2.5), (12, 3.0)])

    context.targets['qtum_btc'] = \
        dict([(1, 1.0), (3, 1.5), (8, 2.0), (30, 2.5), (40, 3.0)])
    #
    # context.targets['lisk_btc'] = \
    #     dict([(50, 1.0), (100, 1.5), (200, 2.0), (400, 2.5), (600, 3.0)])
    #
    # context.targets['xlm_btc'] = \
    #     dict([(10000, 1.0), (20000, 1.5), (40000, 2.0), (80000, 2.5),
    #           (120000, 3.0)])
    #
    # context.targets['stem_btc'] = \
    #     dict([(10, 1.0), (30, 1.5), (80, 2.0), (300, 2.5), (400, 3.0)])
    #
    # context.targets['dcr_btc'] = \
    #     dict([(1, 1.0), (3, 1.5), (8, 2.0), (30, 2.5), (40, 3.0)])


def calculate_orderbook_totals(context, quantity, orderbooks):
    """
    Calculate the cost of the given asset.

    :param context:
    :param symbol:
    :param quantity:
    :param orderbooks:

    :return:
        A dictionary breaking down the cost by:
            exchange_name: str
                The name of the exchange

                    order_type: str
                        The type of orders (bids or asks)
    """
    totals = dict()

    for exchange_name in context.exchanges:
        orderbook = orderbooks[exchange_name]

        totals[exchange_name] = dict()

        if orderbook is None:
            totals[exchange_name]['bids'] = np.nan
            totals[exchange_name]['asks'] = np.nan

        else:
            for order_type in ['bids', 'asks']:
                entries = orderbook[order_type]

                index = 0
                total = 0
                remaining_quantity = quantity

                while remaining_quantity > 0:
                    try:
                        entry = entries[index]

                        entry_quantity = entry['quantity']

                        order_quantity = entry_quantity \
                            if entry_quantity <= remaining_quantity \
                            else remaining_quantity

                        remaining_quantity -= order_quantity

                        total += order_quantity * entry['rate']

                        index += 1
                    except Exception as e:
                        log.warn('order book error on {}, {}'.format(
                            exchange_name, e
                        ))
                        total = np.nan
                        break

                totals[exchange_name][order_type] = total

    return totals


def print_df(df):
    pd.set_option('display.expand_frame_repr', False)
    pd.set_option('precision', 8)
    pd.set_option('display.width', 1000)
    pd.set_option('display.max_colwidth', 1000)

    return df.to_string()


def is_opportunity(context, df, index, percent):
    index = df.index.values[index]

    target = context.targets[index[1]][index[2]]
    if percent > target:
        return True
    else:
        return False


def apply_opportunities(context, df):
    log.info('mapping opportunity signals')
    has_opportunity = False

    for column in df.columns:
        if 'percent' in column[1]:
            label = column[1]

            column_name = label.replace('percent', 'opportunity')

            df[(column[0], column_name)] = \
                [is_opportunity(context, df, idx, df.iloc[idx][column]) for idx
                 in range(len(df))]

            if df[(column[0], column_name)].any():
                has_opportunity = True

    return has_opportunity


def apply_orders(context, df):
    print('applying orders')
    exchanges = context.exchanges.keys()[:]

    while len(exchanges) > 0:
        source_exchange = exchanges.pop(0)

        for target_exchange in exchanges:
            label = '{}_{}'.format(source_exchange, target_exchange)
            order_label = '{}_order'.format(label)

            for ob_type in ['ask', 'bid']:
                opportunity_column = (
                    ob_type, '{}_opportunity'.format(label)
                )
                orders = []

                def calculate_order(row):
                    has_opportunity = row[opportunity_column]

                    value = None
                    if has_opportunity and len(orders) == 0:
                        exchanges = label.split('_')
                        source_exchange = exchanges[0]
                        target_exchange = exchanges[1]

                        if ob_type == 'ask':
                            value = 'buy {}: {}, sell {}: {}'.format(
                                source_exchange, row['ask', source_exchange],
                                target_exchange, row['bid', target_exchange]
                            )

                        elif ob_type == 'bid':
                            value = 'buy {}: {}, sell {}: {}'.format(
                                target_exchange, row['ask', source_exchange],
                                source_exchange, row['bid', target_exchange]
                            )

                        orders.append(value)

                    return value

                sorted_df = df.sort_index(
                    axis=0, level='quantity', ascending=False
                )
                df[ob_type, order_label] = sorted_df.apply(
                    calculate_order, axis=1
                )


def create_orderbook_panel(context, data):
    panel_date = pd.Timestamp.utcnow().floor('1 min')
    panel_hour = panel_date.floor('1H')

    if context.current_hour is None or panel_hour != context.current_hour:
        context.current_hour = panel_hour
        context.hourly_orderbook_data = []

    orderbook_data = []
    for pair_symbol in context.targets:
        orderbooks = dict()
        for exchange_name in context.exchanges:
            try:
                asset = symbol(pair_symbol, exchange_name)

                # TODO: add orderbook api interface
                exchange = context.exchanges[exchange_name]
                orderbooks[exchange_name] = exchange.get_orderbook(asset)

            except SymbolNotFoundOnExchange:
                orderbooks[exchange_name] = None

        for quantity in context.targets[pair_symbol]:

            orderbook_totals = \
                calculate_orderbook_totals(context, quantity, orderbooks)

            for exchange_name in orderbook_totals:
                order_types = orderbook_totals[exchange_name]

                spot_orderbook = dict(
                    date=panel_date,
                    pair_symbol=pair_symbol,
                    quantity=quantity,
                    exchange=exchange_name,
                    ask=order_types['asks'],
                    bid=order_types['bids']
                )

                orderbook_data.append(spot_orderbook)
                context.hourly_orderbook_data.append(spot_orderbook)

    orderbook_df = create_panel(context, orderbook_data)
    has_opportunity = apply_opportunities(context, orderbook_df)

    # TODO: place orders here

    apply_orders(context, orderbook_df)
    result = print_df(orderbook_df)
    print(result)

    if has_opportunity:
        context.num_opportunities += 1

        timestr = panel_date.strftime('%Y%m%d-%H%M%S')
        to_s3(orderbook_df, 'opportunities', timestr)
        send_email(orderbook_df)

    hourly_orderbook_df = create_panel(context, context.hourly_orderbook_data)
    apply_opportunities(context, hourly_orderbook_df)

    timestr = panel_hour.strftime('%Y%m%d-%H')
    to_s3(hourly_orderbook_df, 'hourly', timestr)


def create_panel(context, orderbook_data):
    orderbook_df = pd.DataFrame(orderbook_data)
    orderbook_df.set_index(['date', 'pair_symbol', 'quantity', 'exchange'],
                           inplace=True, drop=True)
    orderbook_df = orderbook_df.unstack(-1)

    for ob_types in [['ask', 'bid'], ['bid', 'ask']]:
        exchanges = context.exchanges.keys()[:]

        while len(exchanges) > 0:
            source_exchange = exchanges.pop(0)

            for target_exchange in exchanges:
                label = '{}_{}'.format(source_exchange, target_exchange)

                if ob_types[0] == 'ask':
                    orderbook_df['ask', label] = \
                        orderbook_df['bid'][target_exchange] - \
                        orderbook_df['ask'][source_exchange]

                    log.info(
                        'buying on {buy_exchange} (ask) selling on '
                        '{sell_exchange} (bid): '
                        '{sell_price} - {buy_price} = {profit}'.format(
                            buy_exchange=source_exchange,
                            sell_exchange=target_exchange,
                            buy_price=
                            orderbook_df['ask'][source_exchange].iloc[0],
                            sell_price=
                            orderbook_df['bid'][target_exchange].iloc[0],
                            profit=orderbook_df['ask', label].iloc[0]
                        )
                    )

                    percent_key = '{}_percent'.format(label)
                    orderbook_df['ask', percent_key] = \
                        orderbook_df['ask'][label] / \
                        orderbook_df['ask'][source_exchange] * 100

                    log.info(
                        'percent profit: {profit} / {buy_price} * 100 = '
                        '{result}'.format(
                            profit=orderbook_df['ask'][label].iloc[0],
                            buy_price=
                            orderbook_df['ask'][source_exchange].iloc[0],
                            result=orderbook_df['ask', percent_key].iloc[0]
                        )
                    )

                else:
                    orderbook_df['bid', label] = \
                        orderbook_df['bid'][source_exchange] - \
                        orderbook_df['ask'][target_exchange]

                    orderbook_df['bid', '{}_percent'.format(label)] = \
                        orderbook_df['bid'][label] / \
                        orderbook_df['ask'][target_exchange] * 100

    return orderbook_df


def send_email(df):
    bytes_to_write = df.to_csv(None).encode()
    return requests.post(
        "https://api.mailgun.net/v3/sandboxcec14dd65c2f4008952e1fad2e79ad19.mailgun.org/messages",
        auth=("api", "key-61013a4d1ff85ddd236a245a6967b72f"),
        files=[
            ("attachment", ("opportunity.csv", bytes_to_write))
        ],
        data={
            "from": "Conatus Alerts <postmaster@conatus.co>",
            "to": ["fredfortier@gmail.com"],
            "subject": "Arbitrage Opportunity",
            "text": "Conatus found the attached opportunity"})


def to_s3(df, subfolder, timestr):
    bytes_to_write = df.to_csv(None).encode()

    obj = s3.Object('conatus-public', 'orderbook-panel/{}/{}-{}.csv'.format(
        subfolder, timestr, algo_namespace
    ))
    obj.put(Body=bytes_to_write)


def create_price_panel(context, data):
    prices = []
    for pair_symbol in context.targets:
        spot_price = dict(
            pair_symbol=pair_symbol,
        )

        for exchange_name in context.exchanges:
            asset = symbol(pair_symbol, exchange_name)
            spot_price[exchange_name] = data.current(asset, 'price')

        prices.append(spot_price)

    price_df = pd.DataFrame(prices)
    price_df.set_index(['pair_symbol'], inplace=True, drop=True)

    exchanges = context.exchanges.keys()[:]
    while len(exchanges) > 0:
        source_exchange = exchanges.pop(0)

        for target_exchange in exchanges:
            label = '{}_{}'.format(source_exchange, target_exchange)
            price_df[label] = \
                price_df[target_exchange] - price_df[source_exchange]

            price_df['{}_percent'.format(label)] = \
                price_df[label] / price_df[source_exchange] * 100

    result = print_df(price_df)
    print(result)

    bytes_to_write = price_df.to_csv(None).encode()
    timestr = time.strftime('%Y%m%d-%H%M%S')

    obj = s3.Object('conatus-public', 'price-panel/{}-{}.csv'.format(
        timestr, algo_namespace
    ))
    obj.put(Body=bytes_to_write)


def create_orders(context, signals):
    log.info('found signal: {}'.format(signals))


def filter_entry_points(context, data):
    """
    Include filters which might stop us from entering positions.

    :param context:
    :param data:
    :return:
    """
    for pair_symbol in context.targets:
        for exchange in context.exchanges.values():
            asset = symbol(pair_symbol, exchange.name)


            # prices = data.history(
            #     asset,
            #     fields='price',
            #     bar_count=30,
            #     frequency='1m'
            # )
            # rsi = talib.RSI(prices.values, timeperiod=14)[-1]
            # print('got rsi: {}'.format(rsi))


def handle_data(context, data):
    exceptions = []
    try:
        # filter_entry_points(context, data)

        # create_price_panel(context, data)

        if context.sim_params.arena != 'backtest':
            create_orderbook_panel(context, data)

    except Exception as e:
        log.warn(e)
        exceptions.append(e)

    log.info('completed bar, exceptions: {}'.format(len(exceptions)))
    record(opportunities=context.num_opportunities)


def analyze(context, stats):
    print('the daily stats:\n{}'.format(get_pretty_stats(stats)))

    pass


# run_algorithm(
#     capital_base=250,
#     start=pd.to_datetime('2017-9-1', utc=True),
#     end=pd.to_datetime('2017-9-28', utc=True),
#     data_frequency='daily',
#     initialize=initialize,
#     handle_data=handle_data,
#     analyze=None,
#     exchange_name='poloniex,bitfinex,bittrex',
#     algo_namespace=algo_namespace,
#     base_currency='btc'
# )
run_algorithm(
    initialize=initialize,
    handle_data=handle_data,
    analyze=analyze,
    exchange_name='poloniex,bittrex',
    live=True,
    algo_namespace=algo_namespace,
    base_currency='btc',
    live_graph=False
)



import os

import pandas as pd
from logbook import Logger
from six import itervalues

from catalyst.constants import LOG_LEVEL
from catalyst.exchange.utils.exchange_utils import get_catalyst_symbol, \
    save_asset_data
from catalyst.exchange.utils.factory import find_exchanges
from catalyst.utils.paths import ensure_directory

log = Logger('TickersCollector', level=LOG_LEVEL)

DATADIR = '/Users/fredfortier/.catalyst-data/exchanges'
ensure_directory(DATADIR)


class TickersCollector:
    def __init__(self, exchanges):
        self.bar_count = 500
        self.exchanges = exchanges

    def __call__(self, exchange_name):
        log.info('scanning tickers for {}'.format(exchange_name))
        exchange = next(
            (exchange for exchange in \
             self.exchanges if exchange.name == exchange_name)
            , None
        )

        try:
            exchange.init()

            exchange_assets = exchange.get_assets()

            tickers = exchange.tickers(exchange_assets)
            all_tickers = list(itervalues(tickers))

            ticks = []
            for ticker in all_tickers:
                ticks.append(dict(
                    ask=ticker['ask'],
                    bid=ticker['bid'],
                    volume=ticker['volume'],
                    last_traded=ticker['last_traded'],
                    symbol=get_catalyst_symbol(ticker['symbol']),
                ))

            df = pd.DataFrame(ticks)
            df.set_index(
                ['symbol', 'last_traded'],
                drop=True,
                inplace=True,
            )
            log.debug('saving tickers')

            now = pd.Timestamp.utcnow()
            timestr = now.strftime('%Y%m')
            exchange_folder = os.path.join(
                DATADIR, exchange.name, timestr, 'tickers'
            )
            ensure_directory(exchange_folder)

            save_asset_data(exchange_folder, df)

        except Exception as e:
            log.warn(
                'unable to fetch tickers for {}: {}'.format(
                    exchange.name, e
                )
            )


if __name__ == '__main__':
    exchanges = find_exchanges(features=['fetchTickers'])
    collector = TickersCollector(exchanges)

    exchange_names = [exchange.name for exchange in exchanges]

    while True:
        for exchange_name in exchange_names:
            collector(exchange_name)



import json
import shutil
import tarfile
import tempfile

import os
import pandas as pd

import boto3
from catalyst.constants import EXCHANGE_CONFIG_URL
from catalyst.data.bundles.core import download_without_progress
from catalyst.exchange.exchange_bcolz import BcolzExchangeBarReader
from catalyst.exchange.utils.exchange_utils import save_exchange_config
from catalyst.exchange.utils.factory import get_exchange, find_exchanges
from catalyst.exchange.utils.serialization_utils import \
    ExchangeJSONDecoder
from catalyst.utils.paths import ensure_directory
from six.moves.urllib import request
from catalyst.assets._assets import TradingPair

s3 = boto3.resource('s3')

ROOT_PATH = '/Users/fredfortier/.catalyst/test'


def get_last_dt_from_file(exchange_name, symbol, freq):
    path = 'catalyst-bundles/exchange-{}/'.format(exchange_name)
    try:
        bucket = s3.Bucket('enigmaco')
        objs = s3.list_objects(
            Bucket='enigmaco',
            Prefix='{}{}-{}-{}'.format(
                path, exchange_name, freq, symbol
            )
        )['Contents']

        objs_m = [o['Key'].decode('utf-8') for o in objs]
        objs_m.sort()
        last_key = objs_m[-1]

        url = 'https://s3.amazonaws.com/enigmaco/{}'.format(last_key)
        bytes = download_without_progress(url)
        path = tempfile.mkdtemp()
        with tarfile.open('r', fileobj=bytes) as tar:
            tar.extractall(path)
        reader = BcolzExchangeBarReader(path)
        last_date = reader.last_available_dt.strftime('%Y-%m-%d')
        shutil.rmtree(path)

    except Exception as e:
        print('unable to find bundles or dates: {}'.format(e))
        last_date = 'N/A'

    return last_date


def update_config_from_bcolz(exchange):
    try:
        url = EXCHANGE_CONFIG_URL.format(exchange=exchange.name)
        handle = request.urlopen(url=url)
        config = json.load(handle, cls=ExchangeJSONDecoder)

        assets = []
        for asset_dict in config['assets']:
            asset = TradingPair(**asset_dict)
            assets.append(asset)

        config['assets'] = assets

    except Exception:
        config = None

    new_config = exchange.create_exchange_config()
    if config is not None:
        new_assets = list(set(new_config['assets']) - set(config['assets']))

        if len(new_assets) > 0:
            for asset in new_assets:
                asset.start_date = pd.Timestamp.utcnow()

                config['assets'].append(asset)

    else:
        config = new_config

    for asset in config['assets']:
        symbol = asset.symbol
        asset.set_end_date(
            get_last_dt_from_file(exchange.name, symbol, 'daily'),
            'daily'
        )
        asset.set_end_date(
            get_last_dt_from_file(exchange.name, symbol, 'minute'),
            'minute'
        )

    folder = os.path.join(ROOT_PATH, 'exchanges', exchange.name)
    ensure_directory(folder)

    filename = os.path.join(folder, 'config.json')
    save_exchange_config(exchange.name, config, filename)

    try:
        s3.upload_file(
            filename,
            'enigmaco',
            'catalyst-exchanges/{}/config.json'.format(exchange.name),
            ExtraArgs={'ACL': 'public-read', 'ContentType': 'application/json'}
        )
    except Exception as e:
        print('unable to upload file: {}'.format(e))


exchanges = find_exchanges()

for exchange in exchanges:
    update_config_from_bcolz(exchange)



import calendar
import os
import tarfile
from datetime import timedelta, datetime, date

import numpy as np
import pandas as pd
import pytz

from catalyst.data.bundles.core import download_without_progress
from catalyst.exchange.utils.exchange_utils import get_exchange_bundles_folder

EXCHANGE_NAMES = ['bitfinex', 'bittrex', 'poloniex']
API_URL = 'http://data.enigma.co/api/v1'


def get_date_from_ms(ms):
    """
    The date from the number of miliseconds from the epoch.

    Parameters
    ----------
    ms: int

    Returns
    -------
    datetime

    """
    return datetime.fromtimestamp(ms / 1000.0)


def get_seconds_from_date(date):
    """
    The number of seconds from the epoch.

    Parameters
    ----------
    date: datetime

    Returns
    -------
    int

    """
    epoch = datetime.utcfromtimestamp(0)
    epoch = epoch.replace(tzinfo=pytz.UTC)

    return int((date - epoch).total_seconds())


def get_bcolz_chunk(exchange_name, symbol, data_frequency, period):
    """
    Download and extract a bcolz bundle.

    Parameters
    ----------
    exchange_name: str
    symbol: str
    data_frequency: str
    period: str

    Returns
    -------
    str
        Filename: bitfinex-daily-neo_eth-2017-10.tar.gz

    """
    root = get_exchange_bundles_folder(exchange_name)
    name = '{exchange}-{frequency}-{symbol}-{period}'.format(
        exchange=exchange_name,
        frequency=data_frequency,
        symbol=symbol,
        period=period
    )
    path = os.path.join(root, name)

    if not os.path.isdir(path):
        url = 'https://s3.amazonaws.com/enigmaco/catalyst-bundles/' \
              'exchange-{exchange}/{name}.tar.gz'.format(
                exchange=exchange_name,
                name=name)

        bytes = download_without_progress(url)
        with tarfile.open('r', fileobj=bytes) as tar:
            tar.extractall(path)

    return path


def get_delta(periods, data_frequency):
    """
    Get a time delta based on the specified data frequency.

    Parameters
    ----------
    periods: int
    data_frequency: str

    Returns
    -------
    timedelta

    """
    return timedelta(minutes=periods) \
        if data_frequency == 'minute' else timedelta(days=periods)


def get_periods_range(start_dt, end_dt, freq):
    """
    Get a date range for the specified parameters.

    Parameters
    ----------
    start_dt: datetime
    end_dt: datetime
    freq: str

    Returns
    -------
    DateTimeIndex

    """
    if freq == 'minute':
        freq = 'T'

    elif freq == 'daily':
        freq = 'D'

    return pd.date_range(start_dt, end_dt, freq=freq)


def get_periods(start_dt, end_dt, freq):
    """
    The number of periods in the specified range.

    Parameters
    ----------
    start_dt: datetime
    end_dt: datetime
    freq: str

    Returns
    -------
    int

    """
    return len(get_periods_range(start_dt, end_dt, freq))


def get_start_dt(end_dt, bar_count, data_frequency, include_first=True):
    """
    The start date based on specified end date and data frequency.

    Parameters
    ----------
    end_dt: datetime
    bar_count: int
    data_frequency: str

    Returns
    -------
    datetime

    """
    periods = bar_count
    if periods > 1:
        delta = get_delta(periods, data_frequency)
        start_dt = end_dt - delta

        if not include_first:
            start_dt += get_delta(1, data_frequency)
    else:
        start_dt = end_dt

    return start_dt


def get_period_label(dt, data_frequency):
    """
    The period label for the specified date and frequency.

    Parameters
    ----------
    dt: datetime
    data_frequency: str

    Returns
    -------
    str

    """
    if data_frequency == 'minute':
        return '{}-{:02d}'.format(dt.year, dt.month)
    else:
        return '{}'.format(dt.year)


def get_month_start_end(dt, first_day=None, last_day=None):
    """
    The first and last day of the month for the specified date.

    Parameters
    ----------
    dt: datetime
    first_day: datetime
    last_day: datetime

    Returns
    -------
    datetime, datetime

    """
    month_range = calendar.monthrange(dt.year, dt.month)

    if first_day:
        month_start = first_day
    else:
        month_start = pd.to_datetime(datetime(
            dt.year, dt.month, 1, 0, 0, 0, 0
        ), utc=True)

    if last_day:
        month_end = last_day
    else:
        month_end = pd.to_datetime(datetime(
            dt.year, dt.month, month_range[1], 23, 59, 0, 0
        ), utc=True)

        if month_end > pd.Timestamp.utcnow():
            month_end = pd.Timestamp.utcnow().floor('1D')

    return month_start, month_end


def get_year_start_end(dt, first_day=None, last_day=None):
    """
    The first and last day of the year for the specified date.

    Parameters
    ----------

    dt: datetime
    first_day: datetime
    last_day: datetime

    Returns
    -------
    datetime, datetime

    """
    year_start = first_day if first_day \
        else pd.to_datetime(date(dt.year, 1, 1), utc=True)
    year_end = last_day if last_day \
        else pd.to_datetime(date(dt.year, 12, 31), utc=True)

    if year_end > pd.Timestamp.utcnow():
        year_end = pd.Timestamp.utcnow().floor('1D')

    return year_start, year_end


def get_df_from_arrays(arrays, periods):
    """
    A DataFrame from the specified OHCLV arrays.

    Parameters
    ----------
    arrays: Object
    periods: DateTimeIndex

    Returns
    -------
    DataFrame

    """
    ohlcv = dict()
    for index, field in enumerate(
            ['open', 'high', 'low', 'close', 'volume']):
        ohlcv[field] = arrays[index].flatten()

    df = pd.DataFrame(
        data=ohlcv,
        index=periods
    )
    return df


def range_in_bundle(asset, start_dt, end_dt, reader):
    """
    Evaluate whether price data of an asset is included has been ingested in
    the exchange bundle for the given date range.

    Parameters
    ----------
    asset: TradingPair
    start_dt: datetime
    end_dt: datetime
    reader: BcolzBarMinuteReader

    Returns
    -------
    bool

    """
    has_data = True
    dates = [start_dt, end_dt]

    while dates and has_data:
        try:
            dt = dates.pop(0)
            close = reader.get_value(asset.sid, dt, 'close')

            if np.isnan(close):
                has_data = False

        except Exception:
            has_data = False

    return has_data


def get_assets(exchange, include_symbols, exclude_symbols):
    """
    Get assets from an exchange, including or excluding the specified
    symbols.

    Parameters
    ----------
    exchange: Exchange
    include_symbols: str
    exclude_symbols: str

    Returns
    -------
    list[TradingPair]

    """
    if include_symbols is not None:
        include_symbols_list = include_symbols.split(',')

        return exchange.get_assets(include_symbols_list)

    else:
        all_assets = exchange.get_assets()

        if exclude_symbols is not None:
            exclude_symbols_list = exclude_symbols.split(',')

            assets = []
            for asset in all_assets:
                if asset.symbol not in exclude_symbols_list:
                    assets.append(asset)

            return assets

        else:
            return all_assets



from catalyst import run_algorithm
from catalyst.api import attach_pipeline, pipeline_output
from catalyst.exchange.exchange_pricing_loader import TradingPairPricing
from catalyst.pipeline import Pipeline, CustomFactor

import pandas as pd

from catalyst.pipeline.factors.equity import SimpleMovingAverage


# Custom factors are defined as a class object, outside of initialize or handle data.
# RecentReturns will calculate the returns for a security over the n-most recent days.
class RecentReturns(CustomFactor):
    # Set the default list of inputs as well as the default window_length.
    # Default values are used if the optional parameters are not specified.
    inputs = [TradingPairPricing.close]
    window_length = 10

    # Computes the returns over the last n days where n = window_length.
    # Any calculation can be performed here and is applied to all markets
    # in the universe.
    def compute(self, today, assets, out, close):
        out[:] = (close[-1] - close[0]) / close[0]


def initialize(context):
    my_pipe = make_pipeline()
    attach_pipeline(my_pipe, 'my_pipeline')

    context.first_frame = True


def make_pipeline():
    """
    Create our pipeline.
    """

    # 10-day close price average.
    mean_10 = SimpleMovingAverage(
        inputs=[TradingPairPricing.close],
        window_length=10,
    )

    # 30-day close price average.
    mean_30 = SimpleMovingAverage(
        inputs=[TradingPairPricing.close],
        window_length=30,
    )

    recent_returns = RecentReturns(
        inputs=[TradingPairPricing.close],
        window_length=14,
    )
    percent_difference = (mean_10 - mean_30) / mean_30

    # Filter to select securities to short.
    shorts = percent_difference.top(5)

    # Filter to select securities to long.
    longs = percent_difference.bottom(5)

    # Filter for all securities that we want to trade.
    securities_to_trade = (shorts | longs)

    pipe = Pipeline(
        columns={
            'longs': longs,
            'shorts': shorts,
            'recent_returns': recent_returns.top(5),
        },
        screen=(securities_to_trade),
    )

    return pipe


def handle_data(context, data):
    if context.first_frame:
        context.output = pipeline_output('my_pipeline')
        context.first_frame = False

    context.longs = context.output[context.output['longs']]
    print(context.longs.index)

    context.shorts = context.output[context.output['shorts']]
    print(context.shorts.index)

    context.recent_returns = context.output[context.output['recent_returns']]
    pass


if __name__ == '__main__':
    mode = 'backtest'

    if mode == 'backtest':
        run_algorithm(
            capital_base=1,
            initialize=initialize,
            handle_data=handle_data,
            analyze=None,
            exchange_name='poloniex',
            algo_namespace='simple_loop',
            base_currency='eth',
            data_frequency='minute',
            start=pd.to_datetime('2017-9-1', utc=True),
            end=pd.to_datetime('2017-12-25', utc=True),
        )
    else:
        run_algorithm(
            capital_base=1,
            initialize=initialize,
            handle_data=handle_data,
            analyze=None,
            exchange_name='binance',
            live=True,
            algo_namespace='simple_loop',
            base_currency='eth',
            live_graph=False,
            simulate_orders=True
        )



import os
from datetime import timedelta
from time import sleep

import pandas as pd
from logbook import Logger

from catalyst.constants import LOG_LEVEL
from catalyst.exchange.exchange_errors import ExchangeRequestError
from catalyst.exchange.utils.bundle_utils import get_start_dt
from catalyst.exchange.utils.factory import find_exchanges
from catalyst.utils.paths import ensure_directory

log = Logger('CandlesCollector', level=LOG_LEVEL)

DATADIR = '/Users/fredfortier/.catalyst-data/exchanges'
ensure_directory(DATADIR)


class CandlesCollector:
    def __init__(self, exchanges):
        self.bar_count = 200
        self.exchanges = exchanges

    def _get_candles(self, exchange, freq, assets, bar_count, start_dt,
                     attempt=0):
        try:
            candles = exchange.get_candles(
                freq=freq,
                assets=assets,
                bar_count=bar_count,
                start_dt=start_dt,
            )
            return candles

        except ExchangeRequestError as e:
            log.warn(
                'unable to fetch candles on attempt {} {}'.format(attempt, e)
            )
            if attempt < 10:
                sleep(5)
                return self._get_candles(
                    exchange=exchange,
                    freq=freq,
                    assets=assets,
                    bar_count=bar_count,
                    start_dt=start_dt,
                    attempt=attempt + 1
                )

            else:
                raise e

    def __call__(self, exchange_name):
        log.info('scanning ohlcv data for {}'.format(exchange_name))
        exchange = next(
            (exchange for exchange in \
             self.exchanges if exchange.name == exchange_name)
            , None
        )

        exchange.init()
        exchange_assets = exchange.get_assets()

        now = pd.Timestamp.utcnow().floor('1T')
        for asset in exchange_assets:
            end_dt = now

            more_bars = True
            while more_bars:
                log.info(
                    'fetching candles {} - {}'.format(
                        asset.symbol, end_dt
                    )
                )
                start_dt = get_start_dt(
                    end_dt, self.bar_count, 'minute'
                )
                try:
                    candles = self._get_candles(
                        exchange=exchange,
                        freq='1T',
                        assets=[asset],
                        bar_count=self.bar_count,
                        start_dt=start_dt,
                    )
                except Exception as e:
                    log.warn('skipping asset on error {}', format(e))
                    break

                # If an asset is no longer available at that date
                # it won't be in the candles collection so we burn it
                if asset not in candles or not candles[asset]:
                    log.info('reached the last candle, moving on '
                             'to the next asset')
                    break

                asset_candles = candles[asset]
                last_candle = asset_candles[-1]
                first_candle = asset_candles[0]

                # Adding a buffer because there are not always trades
                # every minutes. Some exchanges yeild candles only when
                # there is a trade.
                adj_end = end_dt + timedelta(minutes=1)
                if last_candle['last_traded'] > adj_end:
                    log.info('removing duplicate after end date')

                    for candle in asset_candles[:]:
                        if candle['last_traded'] > end_dt:
                            asset_candles.remove(candle)

                    if not asset_candles:
                        break

                adj_start = start_dt + timedelta(minutes=60)
                if first_candle['last_traded'] > adj_start:
                    log.info('start date lated than specified, reached'
                             ' the last candle')

                    more_bars = False

                df = pd.DataFrame(asset_candles)
                df.set_index(
                    'last_traded',
                    drop=True,
                    inplace=True,
                )
                df.sort_index(inplace=True, ascending=False)
                log.debug('saving candles')

                timestr = now.strftime('%Y%m')
                exchange_folder = os.path.join(
                    DATADIR, exchange.name, timestr, 'candles'
                )
                ensure_directory(exchange_folder)

                filename = os.path.join(
                    exchange_folder, '{}.csv'.format(asset.symbol)
                )
                if os.path.exists(filename):
                    print_headers = False

                else:
                    print_headers = True

                with open(filename, 'a') as f:
                    df.to_csv(
                        path_or_buf=f,
                        header=print_headers,
                        float_format='%.{}f'.format(asset.decimals),
                    )

                end_dt = asset_candles[0]['last_traded']


if __name__ == '__main__':
    exchanges = find_exchanges(features=['fetchOHLCV'])
    collector = CandlesCollector(exchanges)

    exchange_names = [exchange.name for exchange in exchanges]

    collector('bittrex')
    # while True:
    #     for exchange_name in exchange_names:
    #         collector(exchange_name)



import os
from web3 import Web3, IPCProvider
import pandas as pd
import json
import numpy as np
from six.moves.urllib import request

pd.set_option('display.height', 1000)
pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)

FIRST_BLOCK = 3154485
LAST_BLOCK = 4700810
INCREMENT_BLOCKS = 100
CONTRACT_ADDRESS = '0x8d12A197cB00D4747a1fe03395095ce2A5CC6819'

contract_json = request.urlopen(
    'https://raw.githubusercontent.com/etherdelta/etherdelta.github.io/master/smart_contract/etherdelta.sol.json'
).read()
json_interface = json.loads(contract_json)

definition_json = request.urlopen(
    'https://raw.githubusercontent.com/etherdelta/etherdelta.github.io/master/config/main.json'
).read()
definition = json.loads(definition_json)


def find_token(addr):
    return next(
        (token for token in definition['tokens'] if
         token['addr'].lower() == addr.lower()),
        dict(name=addr, addr=addr, decimals=18),
    )


def create_trade(event, block):
    args = event['args']
    trade = dict(
        token_get=args['tokenGet'],
        amount_get=np.float64(args['amountGet']),
        token_give=args['tokenGive'],
        amount_give=np.float64(args['amountGive']),
        address_get=args['get'],
        address_give=args['give'],
        block=event['blockNumber'],
        trade_dt=pd.to_datetime(block['timestamp'], unit='s', utc=True),
    )

    if trade['amount_get'] > 0:
        trade['price'] = trade['amount_give'] / trade['amount_get']

    else:
        trade['price'] = 0

    quote_token = find_token(trade['token_give'])
    trade['quote_currency'] = quote_token['name']
    trade['decimals'] = quote_token['decimals']

    base_token = find_token(trade['token_get'])
    trade['base_currency'] = base_token['name']

    if trade['quote_currency'] is not None and trade[
        'base_currency'] is not None:
        trade['symbol'] = '{}_{}'.format(trade['base_currency'].lower(),
                                         trade['quote_currency'].lower())

    return trade


userdir = '/Users/fredfortier'
w3 = Web3(IPCProvider('{}/.ethereum/geth.ipc'.format(userdir)))

token_contract = w3.eth.contract(
    abi=json_interface,
    address=CONTRACT_ADDRESS
)  # Type: Contract

blocks = np.arange(FIRST_BLOCK, LAST_BLOCK, INCREMENT_BLOCKS)
counter = dict()
for index, block in enumerate(blocks):
    try:
        next_block = blocks[index + 1] - 1
    except IndexError:
        print('ingestion complete at block {}'.format(block))
        break

    transfer_filter = token_contract.eventFilter(
        event_name='Trade',
        filter_params=dict(
            fromBlock=int(block),
            toBlock=int(next_block),
        )
    )
    trade_entries = transfer_filter.get_all_entries()

    # deposit_filter = token_contract.eventFilter(
    #     event_name='Deposit',
    #     filter_params=dict(
    #         fromBlock=int(block),
    #         toBlock=int(next_block),
    #     )
    # )
    # deposit_entries = deposit_filter.get_all_entries()
    #
    # withdraw_filter = token_contract.eventFilter(
    #     event_name='Trade',
    #     filter_params=dict(
    #         fromBlock=int(block),
    #         toBlock=int(next_block),
    #     )
    # )
    # withdraw_filter = withdraw_filter.get_all_entries()

    if trade_entries:
        trades = []
        for event in trade_entries:
            block = w3.eth.getBlock(event['blockNumber'])

            trade = create_trade(event, block)
            trades.append(trade)

        if not trades:
            continue

        decimals = trades[0]['decimals']

        df = pd.DataFrame(trades)
        df.set_index(['symbol', 'trade_dt'], inplace=True, drop=True)
        df.sort_index(inplace=True)

        print('block {}:\n{}'.format(event['blockNumber'], df.head()))

        symbols = df.index.get_level_values('symbol')
        for symbol in symbols:
            symbol_df = df.loc[(symbols == symbol)]  # Type: pd.DataFrame
            if symbol in counter:
                counter[symbol] += len(symbol_df.index)

            else:
                counter[symbol] = len(symbol_df.index)

            folder = os.path.join(
                userdir, '.catalyst-data', 'etherdelta', 'trades'
            )
            if not os.path.exists(folder):
                os.makedirs(folder)

            filename = os.path.join(
                folder, '{}.csv'.format(symbol)
            )
            if os.path.exists(filename):
                print_headers = False

            else:
                print_headers = True

            with open(filename, 'a') as f:
                symbol_df.to_csv(
                    path_or_buf=f,
                    header=print_headers,
                    float_format='%.{}f'.format(decimals),
                )

        totals_df = pd.Series(counter)
        totals_filename = os.path.join(folder, 'totals.csv')
        with open(totals_filename, 'w+') as f:
            totals_df.to_csv(f, header=True)

    else:
        print('no trades in range {}-{}'.format(block, next_block))
